@@ -354,14 +354,17 @@ class UnitlabClient:
         try:
             logger.info("Starting Jupyter notebook...")
             
+            # Use NotebookApp for jupyter notebook (not jupyter-server)
             cmd = [
                 "jupyter", "notebook",
                 "--no-browser",
-                "--ServerApp.token=''",
-                "--ServerApp.password=''",
-                "--ServerApp.allow_origin='*'",
-                "--ServerApp.ip='0.0.0.0'",
-                "--ServerApp.port=8888"  # Explicitly use 8888 to match Cloudflare config
+                "--NotebookApp.token=''",
+                "--NotebookApp.password=''", 
+                "--NotebookApp.allow_origin='*'",
+                "--NotebookApp.ip='0.0.0.0'",  # Listen on all interfaces
+                "--NotebookApp.port=8888",
+                "--NotebookApp.allow_root=True",  # Allow root if needed
+                "--NotebookApp.disable_check_xsrf=True"  # For Cloudflare proxy
             ]
             
             self.jupyter_proc = subprocess.Popen(
