@@ -64,6 +64,7 @@ class CloudflareAPITunnel:
         
         self.tunnel_process = None
         self.created_dns_records = []
+        self.tunnel_config_file = None
         
         # Try to initialize binary manager, but don't fail if it doesn't work
         try:
@@ -141,12 +142,12 @@ class CloudflareAPITunnel:
         print(f"🔧 Configuring tunnel routes...")
         
         # Get current tunnel config first
-        get_url = f"{self.api_base}/accounts/{self.account_id}/cfd_tunnel/{self.tunnel_id}/configurations"
+        get_url = f"{self.api_base}/accounts/{self.account_id}/tunnels/{self.tunnel_id}/configurations"
         
         try:
             # Get existing config
             response = requests.get(get_url, headers=self.headers)
-            current_config = response.json()
+            current_config = response.json() if response.status_code == 200 else {}
             
             # Build new ingress rules
             new_ingress = [
@@ -189,7 +190,7 @@ class CloudflareAPITunnel:
                 }
             }
             
-            put_url = f"{self.api_base}/accounts/{self.account_id}/cfd_tunnel/{self.tunnel_id}/configurations"
+            put_url = f"{self.api_base}/accounts/{self.account_id}/tunnels/{self.tunnel_id}/configurations"
             response = requests.put(put_url, headers=self.headers, json=config_data)
             
             if response.status_code == 200:
@@ -205,6 +206,151 @@ class CloudflareAPITunnel:
             print("   Assuming routes are configured in dashboard.")
             return True
 
+    def create_device_tunnel(self):
+        """
+        Create a unique tunnel for this device if it doesn't exist
+        """
+        tunnel_name = f"device-{self.clean_device_id}"
+        print(f"🔍 Checking for tunnel: {tunnel_name}")
+        
+        # Check if tunnel already exists
+        list_url = f"{self.api_base}/accounts/{self.account_id}/tunnels"
+        response = requests.get(list_url, headers=self.headers)
+        
+        if response.status_code == 200:
+            tunnels = response.json().get('result', [])
+            existing_tunnel = None
+            
+            for tunnel in tunnels:
+                if tunnel['name'] == tunnel_name:
+                    existing_tunnel = tunnel
+                    print(f"✅ Found existing tunnel: {tunnel_name}")
+                    break
+            
+            if not existing_tunnel:
+                # Create new tunnel
+                print(f"📦 Creating new tunnel: {tunnel_name}")
+                create_url = f"{self.api_base}/accounts/{self.account_id}/tunnels"
+                # Generate secret first
+                tunnel_secret = os.urandom(32).hex()
+                
+                create_data = {
+                    "name": tunnel_name,
+                    "tunnel_secret": tunnel_secret
+                }
+                
+                create_response = requests.post(create_url, headers=self.headers, json=create_data)
+                
+                if create_response.status_code in [200, 201]:
+                    existing_tunnel = create_response.json()['result']
+                    print(f"✅ Created tunnel: {tunnel_name}")
+                    
+                    # Add the secret to the tunnel info (API doesn't return it)
+                    existing_tunnel['tunnel_secret'] = tunnel_secret
+                    
+                    # Save credentials for this tunnel
+                    self._save_tunnel_credentials(existing_tunnel)
+                    
+                    # Configure tunnel routes
+                    self._configure_tunnel_routes(existing_tunnel['id'])
+                    
+                    # Create DNS records for this device
+                    self.create_dns_records()
+                else:
+                    print(f"❌ Failed to create tunnel: {create_response.text}")
+                    return None
+            else:
+                # Tunnel exists - but we need to ensure it can be used
+                print(f"♻️  Found existing tunnel, setting up for use")
+                
+                # Store tunnel info for later use
+                existing_tunnel['needs_token'] = True  # Mark that we'll need to use token
+                
+                # Configure tunnel routes (creates config file)
+                self._configure_tunnel_routes(existing_tunnel['id'])
+                
+                # Ensure DNS records exist
+                self.create_dns_records()
+            
+            return existing_tunnel
+        
+        return None
+    
+    def _configure_tunnel_routes(self, tunnel_id):
+        """
+        Configure ingress routes for the device tunnel
+        The tunnel needs to be configured with a config file, not via API
+        So we'll create a config file for it
+        """
+        import yaml
+        
+        # Create config file for this tunnel
+        config_dir = Path.home() / '.cloudflared'
+        config_dir.mkdir(exist_ok=True)
+        config_file = config_dir / f'config-{tunnel_id}.yml'
+        
+        config = {
+            "tunnel": tunnel_id,
+            "credentials-file": str(config_dir / f"{tunnel_id}.json"),
+            "ingress": [
+                {
+                    "hostname": f"{self.jupyter_subdomain}.{self.base_domain}",
+                    "service": "http://localhost:8888",
+                    "originRequest": {
+                        "noTLSVerify": True
+                    }
+                },
+                {
+                    "hostname": f"{self.ssh_subdomain}.{self.base_domain}",
+                    "service": "ssh://localhost:22"
+                },
+                {
+                    "service": "http_status:404"
+                }
+            ]
+        }
+        
+        with open(config_file, 'w') as f:
+            yaml.dump(config, f)
+        
+        print(f"✅ Created tunnel config: {config_file}")
+        self.tunnel_config_file = config_file
+    
+    def _save_tunnel_credentials(self, tunnel_info):
+        """
+        Save tunnel credentials locally for this device
+        Credentials must be base64 encoded for cloudflared
+        """
+        import base64
+        import json
+        
+        creds_dir = Path.home() / '.cloudflared'
+        creds_dir.mkdir(exist_ok=True)
+        
+        creds_file = creds_dir / f"{tunnel_info['id']}.json"
+        
+        # Get the secret - it should be hex string
+        secret_hex = tunnel_info.get('tunnel_secret') or tunnel_info.get('secret')
+        if secret_hex:
+            # Convert hex to bytes then to base64
+            secret_bytes = bytes.fromhex(secret_hex)
+            secret_b64 = base64.b64encode(secret_bytes).decode('ascii')
+        else:
+            print("⚠️  No tunnel secret found")
+            return None
+        
+        credentials = {
+            "AccountTag": self.account_id,
+            "TunnelSecret": secret_b64,  # Must be base64!
+            "TunnelID": tunnel_info['id']
+        }
+        
+        with open(creds_file, 'w') as f:
+            json.dump(credentials, f)
+        
+        print(f"💾 Saved credentials to: {creds_file}")
+        return creds_file
+    
     def start_tunnel_with_token(self):
         """
         Start tunnel using the existing service token
@@ -212,28 +358,105 @@ class CloudflareAPITunnel:
         try:
             print("🚀 Starting Cloudflare tunnel...")
             
-            # First, try to set up DNS and routes via API
-            if self.api_token:
-                self.create_dns_records()
-                self.update_tunnel_config()
-            
             # Ensure cloudflared is available
             cloudflared_path = self._ensure_cloudflared()
             if not cloudflared_path:
                 raise RuntimeError("Failed to obtain cloudflared binary")
             
-            # Use service token - simple and reliable
-            # The dashboard must have *.1scan.uz -> localhost:8888 configured
-            service_token = "eyJhIjoiYzkxMTkyYWUyMGE1ZDQzZjY1ZTA4NzU1MGQ4ZGM4OWIiLCJ0IjoiMDc3N2ZjMTAtNDljNC00NzJkLTg2NjEtZjYwZDgwZDYxODRkIiwicyI6Ik9XRTNaak5tTVdVdE1tWTRaUzAwTmpoakxTazBaalF0WXpjek1tSm1ZVGt4WlRRMCJ9"
+            # Create or get existing tunnel for this device
+            device_tunnel = self.create_device_tunnel()
             
-            cmd = [
-                cloudflared_path,
-                "tunnel",
-                "--no-autoupdate",
-                "run",
-                "--token",
-                service_token
-            ]
+            # Now set up DNS and routes via API after tunnel is created/found
+            if self.api_token and device_tunnel:
+                self.tunnel_id = device_tunnel['id']  # Ensure tunnel_id is set
+                self.create_dns_records()
+                self.update_tunnel_config()
+            
+            if not device_tunnel:
+                print("❌ Could not create/find device tunnel")
+                # Fallback to shared tunnel if API fails
+                print("⚠️  Falling back to shared tunnel...")
+                service_token = "eyJhIjoiYzkxMTkyYWUyMGE1ZDQzZjY1ZTA4NzU1MGQ4ZGM4OWIiLCJ0IjoiMDc3N2ZjMTAtNDljNC00NzJkLTg2NjEtZjYwZDgwZDYxODRkIiwicyI6Ik9XRTNaak5tTVdVdE1tWTRaUzAwTmpoakLTazBaalF0WXpjek1tSm1ZVGt4WlRRMCJ9"
+                cmd = [
+                    cloudflared_path,
+                    "tunnel",
+                    "--no-autoupdate",
+                    "run",
+                    "--token",
+                    service_token
+                ]
+            else:
+                tunnel_id = device_tunnel['id']
+                tunnel_name = device_tunnel['name']
+                
+                print(f"🚇 Starting tunnel: {tunnel_name} ({tunnel_id})")
+                
+                # Check if credentials file exists
+                creds_file = Path.home() / '.cloudflared' / f"{tunnel_id}.json"
+                config_file = Path.home() / '.cloudflared' / f'config-{tunnel_id}.yml'
+                cmd = None  # Initialize cmd
+                
+                if not creds_file.exists() or device_tunnel.get('needs_token'):
+                    # Try to recreate credentials from stored secret
+                    if device_tunnel.get('tunnel_secret'):
+                        self._save_tunnel_credentials(device_tunnel)
+                        # Continue to use credentials below
+                    else:
+                        print("⚠️  No stored credentials, using tunnel token...")
+                        # Get token for this tunnel
+                        token_url = f"{self.api_base}/accounts/{self.account_id}/tunnels/{tunnel_id}/token"
+                        token_response = requests.get(token_url, headers=self.headers)
+                        if token_response.status_code == 200:
+                            token = token_response.json()['result']
+                            
+                            # Check if config file exists with ingress rules
+                            if config_file.exists():
+                                # Use token with config file for ingress rules
+                                print(f"   Using token with config file: {config_file}")
+                                cmd = [
+                                    cloudflared_path,
+                                    "tunnel",
+                                    "--no-autoupdate",
+                                    "--config", str(config_file),
+                                    "run",
+                                    "--token", token
+                                ]
+                            else:
+                                # Use token directly
+                                print(f"   Using token directly (no config file)")
+                                cmd = [
+                                    cloudflared_path,
+                                    "tunnel",
+                                    "--no-autoupdate", 
+                                    "run",
+                                    "--token", token
+                                ]
+                        else:
+                            print("❌ Could not get tunnel token")
+                            return None
+                
+                # Only use credentials if we haven't already set cmd with token
+                if cmd is None and creds_file.exists():
+                    # Check if config file exists
+                    if config_file.exists():
+                        # Run tunnel with config file (includes routes)
+                        cmd = [
+                            cloudflared_path,
+                            "tunnel",
+                            "--no-autoupdate",
+                            "--config", str(config_file),
+                            "run"
+                        ]
+                    else:
+                        # Fallback to credentials file only
+                        cmd = [
+                            cloudflared_path,
+                            "tunnel",
+                            "--no-autoupdate",
+                            "--credentials-file", str(creds_file),
+                            "run",
+                            tunnel_id
+                        ]
             
             self.tunnel_process = subprocess.Popen(
                 cmd,
@@ -271,12 +494,17 @@ class CloudflareAPITunnel:
     def stop(self):
         """
         Stop the tunnel if running
+        Note: We keep the tunnel configuration for next run
         """
         if self.tunnel_process and self.tunnel_process.poll() is None:
             print("Stopping tunnel...")
             self.tunnel_process.terminate()
-            self.tunnel_process.wait(timeout=5)
+            try:
+                self.tunnel_process.wait(timeout=5)
+            except subprocess.TimeoutExpired:
+                self.tunnel_process.kill()
             print("Tunnel stopped")
+            print("ℹ️  Tunnel configuration preserved for next run")
 
     def _ensure_cloudflared(self):
         """
